/*  This is a simple backpropagation neural network program. For
    more information regarding the operation and use of this software,
    please read the PBPNN.TXT  file.

    Ardeshir Bahreininejad

   This is the Master task for the parallel version of BPNN software.
   
   This is a single pattern training program where weights are changed 
   after the presentation of one training pattern per Worker.

   This program is distributed in the hope that it will be useful but 
   WITHOUT ANY WARRANTY. 
 
*/


#include  <stdio.h>
#include  <stdlib.h>
#include  <ctype.h>
#include  <string.h>
#include  <signal.h>
#include  <chan.h>
#include  <stddef.h>
#include  <math.h>
#include  <time.h>
#include  <dos.h>

#define    MAX_LINE   300

CHAN    **in_p, **out_p;
int     num_proc;
int     proc_id;

void  set_weights();
void  patterns();
void  training();
void  change_weights();
void  final_weights();
void  test_network();
void  alloc_memory();
void  free_memory();
void  sig_handle();
void  patterns_test();
void  set_net();

/* typedefs and prototypes for dynamic storage of arrays  */

typedef  float       *ptr_FLOAT;
typedef  ptr_FLOAT   VECTOR;
typedef  ptr_FLOAT   *MATRIX;

void  VectorAllocate(VECTOR *vector, int num_cols);
void  AllocateCols(ptr_FLOAT matrix[], int num_rows, int num_cols);
void  MatrixAllocate(MATRIX *pmatrix, int nrows, int ncols);
void  MatrixFree(MATRIX matrix, int num_rows);
int   ERRORCHECK;
int    io_index;
char   sig_flag = 1;
int    p;                     /* index pattern number */
int    h;                     /* index hidden layer */
int    i;                     /* index input layer */
int    j;                     /* index output layer */
float  learn_rate, momentum;  /* tarining parameters */
int    num_io, num_io_test,   /* number of patterns */
       num_input_units,       /* number of input nodes */
       num_hidden_units,      /* number of hidden nodes */
       num_output_units;      /* number of output nodes */

FILE   *fpError;              /* file to save the training errors */
FILE   *fpResults;            /* file to save the results after testing */
FILE   *fpWeights;            /* file to save the weights */
char   fileResults[20];
char   fileWeights[20];
float  wmax, wmin, wrange;
char   buff[MAX_LINE];
int    num_pats;

MATRIX    inputs;                       /* inputs */
MATRIX    net_out_hid;                  /* hidden layer */
MATRIX    delta_inp_hid;                /* delta at hidden layer */
MATRIX    delta_weight_inp_hid;         /* change in weights input:hidden */
MATRIX    weight_inp_hid;               /* weights input:hidden */
MATRIX    net_out_output;               /* output layer */
MATRIX    delta_hid_out;                /* delta at output layer */
MATRIX    delta_weight_hid_out;         /* change in weights hidden:output */
MATRIX    weight_hid_out;               /* weights hidden:output */
MATRIX    target;                       /* target output */

void  main(argc, argv, envp, in_ports, ins, out_ports, outs)
int     argc, ins, outs;
char    *argv[], *envp[];
CHAN    *in_ports[], *out_ports[];
{
void   menu();
char   choice;

    in_p = in_ports;
    out_p = out_ports;


   sprintf(buff, "Enter the number of Workers:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%d", &num_proc);

   proc_id = 0;

   chan_out_word(proc_id, out_p[2]);   /* send processor id to workers */
   chan_out_word(num_proc, out_p[2]);  /* send the number of workers */


   set_net();  /* read the network info from the file with .net */
   num_pats = num_io / num_proc;

   alloc_memory();               /* allocate memory */
   patterns();
   sprintf(buff,"... Training patterns have been setup\n");
   printf(" %s", buff);

   signal(SIGINT, sig_handle);  /* signal handling for interrupting the

   set_weights();  /* setup the weights */

   sprintf(buff,"Enter 1 if you wish to save the errors else 0:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%d", &ERRORCHECK);

   if(ERRORCHECK == 1)
      if ((fpError = fopen("errors.dat", "w")) == NULL)
      {
	  sprintf(buff,"Can not open error-file\n");
	  printf(" %s", buff);
	  exit(1);
      }

   while (1)
   {
      menu();
      sprintf(buff,"Select an option:   ");
      printf(" %s", buff);

      gets(buff);
      sscanf(buff,"%s", &choice);

	switch (choice)
	{
	    case 'o':
	    {
	       menu();
	       break;
	    }
	       case 't':
	    {
	       training();
	       break;
	    }
	       case 's':
	    {
	       final_weights();
	       break;
	    }
	       case 'p':
	    {
	       test_network();
	       break;
	    }
	       case 'q':
	    {
	       if(ERRORCHECK == 1) fclose(fpError);
	       free_memory();
	       exit(1);
	    }
	       default:
	    {
	       sprintf(buff,"Unknown option, try again...\n");
	       printf(" %s", buff);
	    }
	}  /* end of switch */

   }  /* end of while loop */

} /* end main() */


void menu()
{

   sprintf(buff,"\n\n");
   printf(" %s", buff);
   sprintf(buff,"+------------------------------------------------+\n");
   printf(" %s", buff);
   sprintf(buff,"|    A Simple Backpropagation Neutal Net         |\n");
   printf(" %s", buff);
   sprintf(buff,"|                  t -- train a network          |\n");
   printf(" %s", buff);
   sprintf(buff,"|                  s -- save final the weights   |\n");
   printf(" %s", buff);
   sprintf(buff,"|                  p -- test a network           |\n");
   printf(" %s", buff);
   sprintf(buff,"|                  q -- quit program             |\n");
   printf(" %s", buff);
   sprintf(buff,"+------------------------------------------------+\n");
   printf(" %s", buff);

} /* print_menu */


/*  setup weight */

void  set_weights()
{

float   random_weights();
int     max_num_wts;      /* wmax is abs val of max weight */
int     msl_w_inp_hid, msl_w_hid_out ;

   if(num_input_units > num_hidden_units ||
       num_input_units > num_output_units)
				     max_num_wts = num_input_units;
   if(num_hidden_units > num_input_units ||
       num_hidden_units > num_output_units)
				     max_num_wts = num_hidden_units;
   else max_num_wts = num_output_units;

   wmin = 1.0 / 100000.0;
   wmax = 10.0 / 0.5 / (float)max_num_wts;
   wrange = wmax - wmin;

   sprintf(buff,"Weights range is between %f and %f\n", wmin,wmax);
   printf(" %s", buff);

/* set weights to random number between wmin and wmax  */

  msl_w_inp_hid = num_hidden_units * num_input_units+1 * sizeof(float);
  msl_w_hid_out = num_output_units * num_hidden_units+1 * sizeof(float);   

/* read input:hidden weights */

   for  (h = 0; h < num_hidden_units; h++)
   {
     for  (i = 0; i < num_input_units; i++)
     {
	weight_inp_hid[h][i] = random_weights();
	delta_weight_inp_hid[h][i] = 0.0;
     }
   }

/* read hidden:out weights */

   for  (j = 0; j < num_output_units; j++)
   {
     for  (h = 0; h < num_hidden_units; h++)
     {
	weight_hid_out[j][h] = random_weights();
	delta_weight_hid_out[j][h] =  0.0;
     }
   }
   sprintf(buff,"... Weights have been setup for the network\n");
   printf(" %s", buff);

}  /* end of set_weights() */


/* values for random weights */

float  random_weights()
{
float  temp;

   temp = (float)(rand() % 32768) / (float)(1L<<15);
   temp = wmin + wrange * temp;

   if ((rand() % 1000) > 500) temp = temp * -1.0;

   return(temp);

} /* end of random_weights() */


/* read in all patterns to be trained */

void patterns()
{
FILE   *fpPattern;                /* source pattern input file */
char   filePattern[20];
int    num_io_item, msl_item;
float  *io_item;
   
   num_io_item = num_input_units + num_output_units;          
   msl_item = num_io_item * sizeof(float);
   if ((io_item = calloc(num_io_item, sizeof(float)))  ==    NULL)
   {
	sprintf(buff, "Sorry! Not enough memory for vector\n");
	printf(" %s", buff);
	exit(1);
   }

/* read in inputs from the patterns */

   sprintf(buff,"Enter the training pattern filename (.pat extension): ", filePattern);
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%s", filePattern); 
   strcat(filePattern,".pat");

   if  ((fpPattern = fopen(filePattern, "rt")) == NULL)
   {
	sprintf(buff,"Can not open i-o pattern file");
	printf(" %s", buff);
	exit(1);
   }

   for (p = 0; p < num_io; p++)
   {
	for (i = 0;  i < num_io_item;  i++)  
	   fscanf(fpPattern, "%f", &io_item[i]);
	chan_out_message(msl_item, io_item, out_p[2]); 
   }
	     
   free(io_item);
   fclose(fpPattern);

   sprintf(buff,"Total number of patterns is = %d\n", num_io);
   printf(" %s", buff);

}  /* end of patterns() */


/* read in all patterns to be tested */

void patterns_test(fp)
FILE  *fp;
{

/* read in all patterns to be tested */

   for (p = 0; p < num_io_test; p++)
   {
      for (i = 0;  i < num_input_units; i++)
	if (fscanf(fp, "%f", &inputs[p][i]) !=1)
	     goto  ALLPATTERNSREAD;

/* read in target outputs from the patterns */

      for (j = 0;  j < num_output_units;  j++)
	fscanf(fp, "%f", &target[p][j]);
   }

   ALLPATTERNSREAD:

   fclose(fp);
}  /* end of patterns_test() */


/* train the net */

void  training()
{
int       num_cycles=0, num_cycles_show=0, max_cycles=0;
static  int  cycles=0;
float     MAXerr, RMSerr, RMStotal, diff_err;
float     temp_err;
float     targetErr;                 /* error to stop the training */
time_t    tim1, tim2;
int       msl_target, msl_tot_MAXerr, msl_tot_RMSerr;
float     tot_RMSerr, tot_MAXerr;

   msl_target = sizeof(float);
   msl_tot_MAXerr = sizeof(float);
   msl_tot_RMSerr = sizeof(float);

   sprintf(buff,"\n  *****  Press ctrl-c to suspend training ****\n\n");
   printf(" %s", buff);

   sprintf(buff,"Enter target RMS error level:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%f", &targetErr);
   sprintf(buff,"Enter Maximum number of cycles:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%d", &max_cycles);
   sprintf(buff,"Enter number of cycles for showing the results:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%d", &num_cycles_show);
   sprintf(buff,"Enter the learning rate:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%f", &learn_rate);
   sprintf(buff,"Enter the momentum:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%f", &momentum);

   chan_out_message(msl_target, &targetErr, out_p[2]);

   sig_flag = 0;

   RMStotal = targetErr * 2.0;

   time(&tim1);

   while (RMStotal > targetErr)
   {
      if(num_cycles >= max_cycles)
      {
	 sprintf(buff,"\n Network has been trained for %d cycles so far\n",
		  cycles);
	 printf(" %s", buff);
	 break;
      }

      num_cycles++;
      cycles++;

      for (p = 0; p < num_pats; p++)
      {
	sprintf(buff," \b");
	printf(" %s", buff);
	if(sig_flag)  break;
	change_weights();  /* calculate delta-weights and change weights */
   
      } /* end of reading all the patterns */

      chan_in_message(msl_tot_MAXerr, &tot_MAXerr, in_p[2]);
      chan_in_message(msl_tot_RMSerr, &tot_RMSerr, in_p[2]);

      MAXerr = tot_MAXerr;
      RMStotal = sqrt(tot_RMSerr / ((float)num_output_units * (float)num_io));

      /* do not count interrupted cycles */

      if(sig_flag) { cycles--;  break; }

      if(num_cycles_show != 0)
	if ((num_cycles % num_cycles_show) == 0)
	{
	  sprintf(buff,"\nMAXerr = %f  RMSerr = %f",
		       MAXerr,RMStotal);
	  printf(" %s", buff);
	  sprintf(buff,"\nCycle No. = %d", cycles);
	  printf(" %s", buff);
	}

      if(ERRORCHECK == 1)
	 fprintf(fpError,"MAXerr = %f   RMSerr = %f  Cycle No. = %d\n",
		  MAXerr,RMStotal, cycles);

   }  /* end of while */

   time(&tim2);
   sprintf(buff,"\nTotal training time = %f  sec\n", (float)tim2 - (float)tim1);
   printf(" %s", buff);

   if(ERRORCHECK == 1)
   {
       fprintf(fpError,"\n\nTotal training time = %f  sec",
		    (float)tim2 - (float)tim1);
       sprintf(buff,"Errors have been saved in errors.dat file\n");
       printf(" %s", buff);
   }

   if(!sig_flag&&(cycles < max_cycles))
   {
       sprintf(buff,"\n\nNetwork learned after %d cycles\n", cycles);
       printf(" %s", buff);
   }
} /* end of training() */


void  change_weights()
{
int    msl_new_tot_dw, msl_tot_dw;
float  tot_dw, new_tot_dw;

  msl_new_tot_dw = sizeof(float);
  msl_tot_dw = sizeof(float);

/* change weights hidden:output */

  for (j = 0; j < num_output_units; j++)
  {
     for(h = 0; h < num_hidden_units; h++)
     {
       chan_in_message(msl_tot_dw, &tot_dw, in_p[2]);
       new_tot_dw = 0.0;         
       new_tot_dw = (tot_dw * learn_rate)
		   + (momentum * delta_weight_hid_out[j][h]);

       chan_out_message(msl_new_tot_dw, &new_tot_dw, out_p[2]);

/* calculate new weights */

       delta_weight_hid_out[j][h] = new_tot_dw;          
       weight_hid_out[j][h]  += new_tot_dw;

     }
  }

/* change weights input:hidden */
   
   for (h = 0; h < num_hidden_units; h++)
   {
      for (i = 0; i < num_input_units; i++)
      {
	chan_in_message(msl_tot_dw, &tot_dw, in_p[2]);
	new_tot_dw = 0.0;      /* delta weight */   
	new_tot_dw = (learn_rate * tot_dw)
		   + (momentum * delta_weight_inp_hid[h][i]);

       chan_out_message(msl_new_tot_dw, &new_tot_dw, out_p[2]);

  /* calculate new weights */

	weight_inp_hid[h][i] += new_tot_dw;
	delta_weight_inp_hid[h][i] = new_tot_dw;
      }
   }
} /* end change_weights() */



/* save the final weights */

void  final_weights()
{
   if ((fpWeights = fopen("saved.wgt", "w")) == NULL)
   {
	sprintf(buff,"can't write file saved.wgt \n");
	printf(" %s", buff);
	exit(1);
   }

   for (h = 0; h < num_hidden_units; h++)
	for (i = 0; i < num_input_units; i++)
	   fprintf(fpWeights,  "%f\n", weight_inp_hid[h][i]);

   for (j = 0; j < num_output_units; j++)
	for (h = 0;  h < num_hidden_units; h++)
	      fprintf(fpWeights,  "%f\n", weight_hid_out[j][h]);

   sprintf(buff,"\n*** Weights have been saved in file saved.wgt ***\n");
   printf(" %s", buff);

   fclose(fpWeights);

} /*  final_weights()  */


/* test a trained network */

void  test_network()
{
FILE  *fpTest;
char  fileTest[20];

   sprintf(buff,"Enter the saved weights filename -default (saved):  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%s", fileWeights); 
   strcat(fileWeights,".wgt");

   if ((fpWeights = fopen(fileWeights, "r"))  ==  NULL)
   {
	sprintf(buff,"Can not open weights file \n");
	printf(" %s", buff);
	exit(1);
   }

   sprintf(buff,"How many test patterns?:  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%d", &num_io_test);

   sprintf(buff,"Enter test filename (.tst extension):  ");
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%s", fileTest); 
   strcat(fileTest,".tst");

   if  ((fpTest = fopen(fileTest, "r")) == NULL)
   {
	sprintf(buff,"Can not open i-o pattern test file %s", fileTest);
	printf(" %s", buff);
	exit(1);
   }

   patterns_test(fpTest);
   sprintf(buff,"...  %d patterns have been setup for testing\n\n",
	       num_io_test);
   printf(" %s", buff);

/* save the rsults in a file with extension .res */

    sprintf(buff,"Enter the filename for saving results (.res assumed):  ");
    printf(" %s", buff);
    gets(buff);
    sscanf(buff,"%s", fileResults); 
    strcat(fileResults,".res");

   if ((fpResults = fopen(fileResults, "w"))  ==  NULL)
   {
	sprintf(buff,"can't open file results file \n");
	printf(" %s", buff);
	exit(1);
   }


/* read and setup weights */

   for  (h = 0; h < num_hidden_units; h++)
     for  (i = 0; i < num_input_units; i++)
	fscanf(fpWeights, "%f\n", &weight_inp_hid[h][i]);


   for  (j = 0; j < num_output_units; j++)
     for  (h = 0; h < num_hidden_units; h++)
	fscanf(fpWeights, "%f\n", &weight_hid_out[j][h]);

   for (p = 0; p < num_io_test; p++)
   {

/* propagate inputs to hidden layer  */

      for (h = 0; h < num_hidden_units; h++)
      {
	  float sum = 0.0;
	  for (i = 0; i < num_input_units; i++)
	      sum += weight_inp_hid[h][i] * inputs[p][i];

/* compute output (use sigmold) */

	  net_out_hid[p][h] = 1.0 / (1.0 + exp(-1.0 * sum));
      }

/* propagate to output layer */

      for (j = 0; j < num_output_units; j++)
      {
	  float  sum = 0.0;
	  for (h = 0; h < num_hidden_units; h++)
		sum += weight_hid_out[j][h] * net_out_hid[p][h];

	  net_out_output[p][j] = 1.0 / (1.0 + exp(-1.0 * sum));
	  fprintf(fpResults,"The outputs of pattern %d is  %f\n",
			  p+1, net_out_output[p][j]);
      }

   }

   sprintf(buff,"*** The test results have been saved in file results.dat ***\n");
   printf(" %s", buff);

   fclose(fpWeights);
   fclose(fpResults);

} /* end test_network() */


/* allocate memory */

void  alloc_memory()
{
   MatrixAllocate(&inputs, num_io, num_input_units);
   MatrixAllocate(&net_out_hid, num_io, num_hidden_units);
   MatrixAllocate(&net_out_output, num_io, num_output_units);
   MatrixAllocate(&delta_hid_out, num_io, num_output_units);
   MatrixAllocate(&delta_weight_hid_out, num_output_units, num_hidden_units+1);
   MatrixAllocate(&weight_hid_out, num_output_units, num_hidden_units+1);
   MatrixAllocate(&delta_inp_hid, num_io, num_hidden_units);
   MatrixAllocate(&delta_weight_inp_hid, num_hidden_units, num_input_units+1);
   MatrixAllocate(&weight_inp_hid, num_hidden_units, num_input_units+1);
   MatrixAllocate(&target, num_io, num_output_units);

} /* end alloc_memory() */


/* free memory */

void  free_memory()
{
 
    MatrixFree(inputs, num_io);
    MatrixFree(net_out_hid, num_io);
    MatrixFree(delta_inp_hid, num_io);
    MatrixFree(delta_weight_inp_hid, num_hidden_units);
    MatrixFree(weight_inp_hid, num_hidden_units);
    MatrixFree(net_out_output, num_io);
    MatrixFree(delta_hid_out, num_io);
    MatrixFree(delta_weight_hid_out, num_output_units);
    MatrixFree(weight_hid_out, num_output_units);
    MatrixFree(target, num_io);
}  /* end free_memory() */


/* Array storage allocation routines */

void  VectorAllocate(VECTOR *vector, int num_cols)
{
    if ((*vector = (VECTOR) calloc(num_cols, sizeof(float))) == NULL)
    {
	sprintf(buff,"Sorry! Not enough memory for nodes\n");
	printf(" %s", buff);
	exit(1);
    }
} /* end of VectorAllocate() */


void AllocateCols(ptr_FLOAT matrix[], int num_rows, int num_cols)
{
   int  i;

   for (i = 0;  i < num_rows;  i++)
	VectorAllocate(&matrix[i], num_cols);
} /* end of AllocateCOls()  */


/* Allocate space for a two dimensional dynamic matrix */

void MatrixAllocate(MATRIX *pmatrix, int num_rows, int num_cols)
{
   if ((*pmatrix  = (MATRIX) calloc(num_rows,  sizeof(ptr_FLOAT)))  ==    NULL)
   {
	sprintf(buff, "Sorry! Not enough memory for nodes\n");
	printf(" %s", buff);
	exit(1);
   }

   AllocateCols(*pmatrix, num_rows, num_cols);
} /*  end of MatrixAllocate()  */


/* free space */

void MatrixFree(MATRIX matrix,  int num_rows)
{
   int   i;
   for (i = 0;  i < num_rows;  i++) free(matrix[i]);

   free(matrix);
}  /* end ofMatrixFree()  */


void sig_handle()
{

    signal (SIGINT, sig_handle);
    sig_flag = 1;

} /* end of sig_handle() */


void  set_net()
{
FILE   *fpNet;                /* file with the network information */
char   fileNet[20];
int    msl_data;
int    *data;

   sprintf(buff,"Enter the network filename (.net extension):  ", fileNet);
   printf(" %s", buff);
   gets(buff);
   sscanf(buff,"%s", fileNet); 
   strcat(fileNet,".net");

   if  ((fpNet = fopen(fileNet, "rt")) == NULL)
   {
	sprintf(buff,"Can not open network file %s", fileNet);
	printf(" %s", buff);
	exit(1);
   }

     fseek(fpNet, 0, SEEK_SET);
     fscanf(fpNet,"%d  %d  %d  %d", &num_input_units, &num_hidden_units,
				     &num_output_units, &num_io);
       if((data = calloc(4,sizeof(int))) == NULL)
	 {printf("out of memory for data[]\n"); exit(1);}
	
	msl_data = 4*sizeof(int);
	data[0] = num_input_units;
	data[1] = num_hidden_units;
	data[2] = num_output_units;
	data[3] = num_io;
	chan_out_message(msl_data, data, out_p[2]);  
     free(data);
     fclose(fpNet);

}
