/*  This is a simple backpropagation neural network program. For
    more information regarding the operation and use of this software,
    please read the PBPNN.TXT file.

    Ardeshir Bahreininejad

    This is the Worker task program for the parallel BPNN software.
    
    This is a single pattern training program where weights are changed 
    after the presentation of one training pattern per Worker.

    This program is distributed in the hope that it will be useful but 
    WITHOUT ANY WARRANTY. 

*/

#include  <stdio.h>
#include  <stdlib.h>
#include  <ctype.h>
#include  <string.h>
#include  <signal.h>
#include  <chan.h>
#include  <stddef.h>
#include  <math.h>
#include  <time.h>
#include  <dos.h>


CHAN    **in_p, **out_p;
int     num_proc;
int     proc_id;

void  set_weights();
void  patterns();
void  training();
void  prop_forward();
void  delta_signals();
void  change_weights();
void  final_weights();
void  alloc_memory();
void  free_memory();
void  set_net();

/* typedefs and prototypes for dynamic storage of arrays  */

typedef  float       *ptr_FLOAT;
typedef  ptr_FLOAT   VECTOR;
typedef  ptr_FLOAT   *MATRIX;

void  VectorAllocate(VECTOR *vector, int num_cols);
void  AllocateCols(ptr_FLOAT matrix[], int num_rows, int num_cols);
void  MatrixAllocate(MATRIX *pmatrix, int nrows, int ncols);
void  MatrixFree(MATRIX matrix, int num_rows);
int   ERRORCHECK;
int    p;                     /* index pattern number */
int    h;                     /* index hidden layer */
int    i;                     /* index input layer */
int    j;                     /* index output layer */
int    num_io,                /* number of patterns */
       num_input_units,       /* number of input nodes */
       num_hidden_units,      /* number of hidden nodes */
       num_output_units;      /* number of output nodes */
int    num_pats;

MATRIX    inputs;                       /* inputs */
MATRIX    net_out_hid;                  /* hidden layer */
MATRIX    delta_inp_hid;                /* delta at hidden layer */
MATRIX    delta_weight_inp_hid;         /* change in weights input:hidden */
MATRIX    weight_inp_hid;               /* weights input:hidden */
MATRIX    net_out_output;               /* output layer */
MATRIX    delta_hid_out;                /* delta at output layer */
MATRIX    delta_weight_hid_out;         /* change in weights hidden:output */
MATRIX    weight_hid_out;               /* weights hidden:output */
MATRIX    target;                       /* target output */
float     wmin, wmax, wrange;  

void  main(argc, argv, envp, in_ports, ins, out_ports, outs)
int     argc, ins, outs;
char    *argv[], *envp[];
CHAN    *in_ports[], *out_ports[];
{
void   menu();
char   choice;

    in_p = in_ports;
    out_p = out_ports;
    
    /* receive worker id's and number of workers */

    chan_in_word(&proc_id, in_p[1]);
    proc_id++; 
    chan_in_word(&num_proc, in_p[1]);

    if(proc_id != num_proc)
    {
       chan_out_word(proc_id, out_p[0]);
       chan_out_word(num_proc, out_p[0]);
    }
   

/* receive info for the network from the Master */  
   set_net();

   num_pats = num_io / num_proc;   
   alloc_memory();               /* allocate memory */
   
 

   patterns();   /* receive the training patterns from the Master */

   set_weights(); /* set weights */
   training();    /* train the net */

} /* end main() */


/*  setup weight */

void  set_weights()
{
float   random_weights();
int     max_num_wts;      /* wmax is abs val of max weight */


   if(num_input_units > num_hidden_units ||
       num_input_units > num_output_units)
				     max_num_wts = num_input_units;
   if(num_hidden_units > num_input_units ||
       num_hidden_units > num_output_units)
				     max_num_wts = num_hidden_units;
   else max_num_wts = num_output_units;

   wmin = 1.0 / 100000.0;
   wmax = 10.0 / 0.5 / (float)max_num_wts;
   wrange = wmax - wmin;

/* set weights to random number between wmin and wmax  */

/* read input:hidden weights */

   for  (h = 0; h < num_hidden_units; h++)
   {
     for  (i = 0; i < num_input_units; i++)
     {
	weight_inp_hid[h][i] = random_weights();
	delta_weight_inp_hid[h][i] = 0.0;
     }
   }
/* read hidden:out weights */

   for  (j = 0; j < num_output_units; j++)
   {
     for  (h = 0; h < num_hidden_units; h++)
     {
	weight_hid_out[j][h] = random_weights();
	delta_weight_hid_out[j][h] =  0.0;
     }
   }

}  /* end of set_weights() */


/* values for random weights */

float  random_weights()
{
float  temp;

   temp = (float)(rand() % 32768) / (float)(1L<<15);
   temp = wmin + wrange * temp;

   if ((rand() % 1000) > 500) temp = temp * -1.0;

   return(temp);

} /* end of random_weights() */


/* read in all patterns to be trained */

void patterns()
{
int     m, k, msl_item, rest_pats;
float   *io_item;     
int     num_io_item;
   
   num_io_item = num_input_units + num_output_units;      
   rest_pats = num_io - proc_id * num_pats;
   msl_item = num_io_item * sizeof(float);
   io_item = calloc(num_io_item, sizeof(float));

   if(proc_id == num_proc)
   {
       for(p = 0; p < num_pats; p++)
       {
	  chan_in_message(msl_item, io_item, in_p[1]);   
	  for (i = 0; i < num_input_units; i++)       
	      inputs[p][i] = io_item[i];    
	  for (j = num_input_units, k = 0;  j < num_io_item, k < num_output_units;  j++, k++)    
	      target[p][k] = io_item[j];    
       }
   }
   else
   {
       for(m = 0; m < rest_pats; m++)
       {
	   chan_in_message(msl_item, io_item, in_p[1]);  
	   chan_out_message(msl_item, io_item, out_p[0]);  
       }     
       for(p = 0; p < num_pats; p++)
       {
	   chan_in_message(msl_item, io_item, in_p[1]);   
	   for (i = 0;  i < num_input_units; i++)       
	       inputs[p][i] = io_item[i];    
	   for (j = num_input_units, k = 0;  j < num_io_item, k < num_output_units;  j++, k++)    
	       target[p][k] = io_item[j];    
       }
   }        

      free(io_item);

}  /* end of patterns() */


/* train the net */

void  training()
{
int      msl_target, msl_tot_MAXerr, msl_tot_RMSerr;
float     MAXerr, RMSerr, RMStotal, diff_err, tot_RMSerr, tot_MAXerr;
float     temp_err;
float     targetErr;                 /* error to stop the training */

   msl_target = sizeof(float);
   msl_tot_MAXerr = sizeof(float);
   msl_tot_RMSerr = sizeof(float);

   if(proc_id == num_proc)
       chan_in_message(msl_target, &targetErr, in_p[1]);
   else
   {
       chan_in_message(msl_target, &targetErr, in_p[1]);
       chan_out_message(msl_target, &targetErr, out_p[0]);
   }
   
   RMStotal = targetErr * 2.0;

   while(1)
   {
   temp_err = 0.0;
   diff_err = 0.0;
   RMSerr = tot_RMSerr = 0.0;
   MAXerr = tot_MAXerr = 0.0;
      
      for (p = 0; p < num_pats; p++)
      {
	 prop_forward();   /* forward propagation of inputs */

	/* keep a check on error difference between net output
	 and target output. Also calculate the Square of errors for
	 calculating the RMS later*/

	 for (j = 0;  j < num_output_units;  j++)
	 {
	    temp_err = fabs(target[p][j] - net_out_output[p][j]);
	    RMSerr += temp_err * temp_err;
	 }

	 delta_signals();  /* calculate delta signals */
	 
	 change_weights();  /* calculate delta-weights and change weights */
	    
	 if(temp_err > diff_err) diff_err = temp_err;

      } /* end of reading all the patterns */

      MAXerr = diff_err;
      RMStotal = RMSerr;

      if(proc_id == num_proc)
      {
	 tot_MAXerr = MAXerr;
	 tot_RMSerr = RMStotal;
	 chan_out_message(msl_tot_MAXerr, &tot_MAXerr, out_p[1]);
	 chan_out_message(msl_tot_RMSerr, &tot_RMSerr, out_p[1]);
      }
      else 
      {
	 chan_in_message(msl_tot_MAXerr, &tot_MAXerr, in_p[0]);
	 chan_in_message(msl_tot_RMSerr, &tot_RMSerr, in_p[0]);
	 if(MAXerr > tot_MAXerr)   
	 {
	    tot_MAXerr = MAXerr;
	    chan_out_message(msl_tot_MAXerr, &tot_MAXerr, out_p[1]);
	 }
	 else 
	      chan_out_message(msl_tot_MAXerr , &tot_MAXerr, out_p[1]);
	 
	 tot_RMSerr += RMStotal;
	 chan_out_message(msl_tot_RMSerr, &tot_RMSerr, out_p[1]);
      }

   }  /* end of while */

} /* end of training() */



void  prop_forward()
{
float  sum;

/* propagate inputs to hidden layer */

	 for (h = 0; h < num_hidden_units; h++)
	 {
	     sum = 0.0;
	     for (i = 0; i < num_input_units; i++)
		 sum += weight_inp_hid[h][i] * inputs[p][i];

/* compute output (use sigmold) */

	     net_out_hid[p][h] = 1.0 / (1.0 + exp(-1.0 * sum));
	 }

/* propagate to output layer */

	 for (j = 0; j < num_output_units; j++)
	 {
	     sum = 0.0;
	     for (h = 0; h < num_hidden_units; h++)
		sum += weight_hid_out[j][h]  *  net_out_hid[p][h];

	     net_out_output[p][j] = 1.0 / (1.0 + exp(-1.0 * sum));
	 }

} /* end prop_forward() */


void   delta_signals()
{
float   sum;

/* compute delta signals for each output unit for a given pattern */

	 for (j = 0; j < num_output_units; j++)
	    delta_hid_out[p][j] = (target[p][j] - net_out_output[p][j]) *
		     net_out_output[p][j] * (1.0 - net_out_output[p][j]);

/* delta signals for hidden units */

	 for (h = 0; h < num_hidden_units; h++)
	 {
	    sum = 0.0;
	    for (j = 0; j < num_output_units; j++)
		sum += delta_hid_out[p][j] * weight_hid_out[j][h];

	    delta_inp_hid[p][h] = sum * net_out_hid[p][h] *
					 (1.0 - net_out_hid[p][h]);
	 }

}  /* end delat_signals() */


void  change_weights()
{
int   msl_new_tot_dw, msl_tot_dw;
float  tot_dw, new_tot_dw;
float  dw, temp;

      msl_new_tot_dw = sizeof(float);
      msl_tot_dw = sizeof(float);

/* change weights hidden:output */

  for (j = 0; j < num_output_units; j++)
  {
     dw = 0.0;                     /* delta weight */
     temp = delta_hid_out[p][j];
     for(h = 0; h < num_hidden_units; h++)
     {
       dw = (temp * net_out_hid[p][h]);
       if(proc_id == num_proc)
       {
	   tot_dw = dw;
	   chan_out_message(msl_tot_dw, &tot_dw, out_p[1]);
	   chan_in_message(msl_new_tot_dw , &new_tot_dw, in_p[1]);
       }
       else
       {
	  chan_in_message(msl_tot_dw, &tot_dw, in_p[0]);
	  tot_dw += dw;
	  chan_out_message(msl_tot_dw, &tot_dw, out_p[1]);
	  chan_in_message(msl_new_tot_dw , &new_tot_dw, in_p[1]);
	  chan_out_message(msl_new_tot_dw , &new_tot_dw, out_p[0]);
       }
       weight_hid_out[j][h]  += new_tot_dw;
       delta_weight_hid_out[j][h] = new_tot_dw;
     }  /* end inner for loop */
  }    /* end outer for loop */

/* change weights input:hidden */
   for (h = 0; h < num_hidden_units; h++)
   {
      dw = 0.0;                  /* delta weight */
      temp = delta_inp_hid[p][h];
      for (i = 0; i < num_input_units; i++)
      {
       dw = (temp * inputs[p][i]);
       if(proc_id == num_proc)
       {
	   tot_dw = dw;
	   chan_out_message(msl_tot_dw, &tot_dw, out_p[1]);
	   chan_in_message(msl_new_tot_dw , &new_tot_dw, in_p[1]);
       }
       else
       {
	  chan_in_message(msl_tot_dw, &tot_dw, in_p[0]);
	  tot_dw += dw;
	  chan_out_message(msl_tot_dw, &tot_dw, out_p[1]);
	  chan_in_message(msl_new_tot_dw , &new_tot_dw, in_p[1]);
	  chan_out_message(msl_new_tot_dw , &new_tot_dw, out_p[0]);
       }
	weight_inp_hid[h][i] += new_tot_dw;
	delta_weight_inp_hid[h][i] = new_tot_dw;
      }
   }
} /* end change_weights() */


/* allocate memory */

void  alloc_memory()
{
   MatrixAllocate(&inputs, num_pats, num_input_units);
   MatrixAllocate(&net_out_hid, num_pats, num_hidden_units);
   MatrixAllocate(&net_out_output, num_pats, num_output_units);
   MatrixAllocate(&delta_hid_out, num_pats, num_output_units);
   MatrixAllocate(&delta_weight_hid_out, num_output_units, num_hidden_units);
   MatrixAllocate(&weight_hid_out, num_output_units, num_hidden_units);
   MatrixAllocate(&delta_inp_hid, num_pats, num_hidden_units);
   MatrixAllocate(&delta_weight_inp_hid, num_hidden_units, num_input_units);
   MatrixAllocate(&weight_inp_hid, num_hidden_units, num_input_units);
   MatrixAllocate(&target, num_pats, num_output_units);

} /* end alloc_memory() */


/* free memory */

void  free_memory()
{
    MatrixFree(inputs, num_pats);
    MatrixFree(net_out_hid, num_pats);
    MatrixFree(delta_inp_hid, num_pats);
    MatrixFree(delta_weight_inp_hid, num_hidden_units);
    MatrixFree(weight_inp_hid, num_hidden_units);
    MatrixFree(net_out_output, num_pats);
    MatrixFree(delta_hid_out, num_pats);
    MatrixFree(delta_weight_hid_out, num_output_units);
    MatrixFree(weight_hid_out, num_output_units);
    MatrixFree(target, num_pats);

}  /* end free_memory() */


/* Array storage allocation routines */

void  VectorAllocate(VECTOR *vector, int num_cols)
{
    *vector = (VECTOR) calloc(num_cols, sizeof(float));

} /* end of VectorAllocate() */


void AllocateCols(ptr_FLOAT matrix[], int num_rows, int num_cols)
{
   int  i;

   for (i = 0; i < num_rows; i++)
	VectorAllocate(&matrix[i], num_cols);

} /* end of AllocateCOls()  */


/* Allocate space for a two dimensional dynamic matrix */

void MatrixAllocate(MATRIX *pmatrix, int num_rows, int num_cols)
{
   *pmatrix  = (MATRIX) calloc(num_rows,  sizeof(ptr_FLOAT));
   AllocateCols(*pmatrix, num_rows, num_cols);

} /*  end of MatrixAllocate()  */


/* free space */

void MatrixFree(MATRIX matrix,  int num_rows)
{
   int   i;
   for (i = 0;  i < num_rows;  i++) free(matrix[i]);

   free(matrix);
}  /* end ofMatrixFree()  */

void  set_net()
{
int     msl_data; 
int    *data;
	
	data = calloc(4,sizeof(int));
	msl_data = 4*sizeof(float);
       if(proc_id == num_proc)
	  chan_in_message(msl_data, data, in_p[1]);  
       else   
       {
	  chan_in_message(msl_data, data, in_p[1]);
	  chan_out_message(msl_data, data, out_p[0]);
       }
	
	  num_input_units = data[0];
	  num_hidden_units = data[1];
	  num_output_units = data[2];
	  num_io = data[3];
      
      free(data);
}


