/* USES MEAN FIELD ANNEALING CODE FOR SUB-DOMAIN GENERATION
   This program uses the Mean Field Anealing NN to optimize the partition
   of a FE mesh.  If choice is 0 then mesh is partitioned without the
   predictive model, if choice is 1 the mesh is partitioned using the
   predictive mode  (See the following paper)

   Subdomain Generation Using Multiple Neural Networks Models
   A. Bahreininejad, B. H. V. Topping and A. I. Khan,
   in the proceedings of 1994 Computational Structures Technology Conference

   Maximum number of predicted elements within an element is 103.
   This software may be compiled for PC and a single transputer.


*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <dos.h>

#define MAX_LINE_SIZE 256  /* for input */

#define NEsub  300
#define NNsub  200
#define NSGT   200
#define LIL 2
#define MM  15
#define NC  20
#define LASTNODE 2000   /* take up to last node in the .dat file */
#define UnBalance 20;   /*used in form_decomp_dec */

#define Pi        3.141592654

int     seed=5;


int     nn, ne, *nod, choice;
int     ne1, *mark1, ne2, *mark2;
double  *cord;
double  *hi, *hhi, *cg, *area, *err;
double  left_val,right_val,total_val;
int     div_hv;
double  xseg1, yseg1, xseg2, yseg2;
int     bsgt;
int     *seg;
int     t_sds, *ne_g;          /*to trace global elnos*/
int     *elems_sd, *elem_sdno;
int     nb,nm,nl,ndime,nstr,nbe,kmax,nmax,iln,nq;

int     nodl[LIL][3],nodq[LIL][5],
	nco[MM][4],nsnod[LIL][5],
	mat[5],nc[NC],ngs[LIL];

char    titel[50];
double  dt,damp;
double  amat[50],aload[NC*3];

typedef struct{
  int nn,ne,nb,nm,nl,ndime,nstr,nbe,kmax,nmax,iln,nq;
  int nod[4*NEsub],nodl[LIL][3],nodq[LIL][5],
      nco[MM][4],nsnod[LIL][5], mat[5],nc[5],ngs[LIL];
  int ne_g[NEsub]; /*global elnos for the subdoamin*/
  double hhi[NNsub], err[NEsub], hi[NEsub];
  char titel[50];
  double dt,damp;
  double cord[3*NNsub];
  double amat[50],aload[NC/2*3];
  } sub_t;
sub_t SUB;

double   *v;
float    *s;    /* Matrix t (weights), vector v for annealing and */
int      *t;          /* vector s for the partitioned result */
int      mfa_it;   /* iterations during annealing  */


int     *sort;
void    rinput(FILE *);
void    rindec(int it, FILE *fp3);
void    rinmen(FILE *fp5);
void    routput(FILE *);
void    routdec(int nex, FILE *fp4);
void    routmen(int nnx, int nex, FILE *fp6);
void    form_boun(void);
void    compare_and_stack(int i);
void    stack_seg(int nd1, int nd2);
void    stack(int elem_g);
int     comp_st(double x, double y, double z);
void    revise_nod(int trace[], int *pjj, int mat_no);
void    nb_iln(void);
void    partition(FILE *, FILE *, FILE *);
void    mfa(FILE *fpn, float tmp, float alpha);
void    copy(char *arg1, char *arg2);
void    init_idec(int *ne_orig);
void    node_para_err(char *file_sav);
double  neuron(double x[], double y[], double del[] );
void    search_node(void);
void    form_decomp_dec(int ne_orig);
void    load_elem_sdno(register i, int ne_sd);
void    rinput_init(char *file);
void    filer(int flag, int choice, FILE **fp, FILE **fp2, FILE **fp3, FILE **fp4,
	   FILE **fp5, FILE **fp6);
void    elemcord_del(int n,int nnn[], double x[], double y[], double del[]);
void    nodecord_del(int node, double *px, double *py, double *pdel);
void    seglen(double x1,double y1, double x2, double y2, double *pL);
void    calcang(double x[], double y[], double L[], double A[]);
void    dot(double x1, double y1, double x2, double y2,
	 double x3, double y3, double x4, double y4, double *pD);
int     hoex15(double Yin[9], double Yout[1] );
int     side21( double Yin[], double Yout[]);
void    homogen_load(double *L, double *A, double *del, double *Yin);
void    homogen_load_2(double *L, double *A, double *del, double *Yin);

/* these functions are used for mean field annealing */

void    sort_parts();
void    divider_1();
void    weight_matrix();
void    set_v();
void    anneal(float tmp, float alpha);
void    outfile_mfa(FILE *fpn, float tot_pred_elems, float sum_penalty,
	 float penalty, float cost, int numface, float energy);

/* just some printing and clock functions */

void    IO_print();
float   IO_my_get_float();
int     IO_my_get_int();
float   get_time();



static char in_string[MAX_LINE_SIZE];  /* used for 'my' IO routines   */
					  /* and IO_get_float_from_file  */

char IO_str[MAX_LINE_SIZE];     /* two strings for output      */
char IO_wkstr[MAX_LINE_SIZE];

main()
{
FILE *fp,*fp2,*fp3,*fp4, *fp5, *fp6, *fpn;
char c, filename[30], filename_sav[30];
register i,j,it,rc;
int flag;
int nrec,nsd;
int ne_orig;
float  t1, tmp, alpha; /* tmp is the temperature and alpha is imbalance*/
static  float  tim=0.0;

  clrscr();

  printf("		FINITE ELEMENT MESH PARTITIONING\n");
  printf("			      USING\n");
  printf("		   MEAN FIELD NEURAL NETWORKS\n");
  printf("\n\n\n");
  printf("Enter the input data-file name (without extension)\n");
  gets(filename);
  strcpy(filename_sav,filename);
  strcat(filename,".dat");
  copy(filename,"i.dat");

  printf("Enter [1] for predicted elements decompostion and [0] else\n");
  gets(filename); sscanf(filename,"%d",&choice);

  printf("The number of subdomains to be formed N = 2^(no_of_recursions)\n");
  sprintf(IO_str,"Enter the no_of_recursions\n");
  IO_print(0);
  nrec = IO_my_get_int();

/* The information about the state of annealing and the neurons are
   written in file mfa.out */

  if ((fpn = fopen("mfa.out","w")) == NULL)
      {printf("Error opening mfa.out!!!\n"); exit(1);}


  printf("\n");
  sprintf(IO_str,"Enter the initial temperature:\t");
  IO_print(0);
  tmp = IO_my_get_float();
  sprintf(IO_str,"Enter the imbalance factor alpha:\t");
  IO_print(0);
  alpha = IO_my_get_float();


  t_sds = (int) pow(2.,nrec); /* total no. of subdomains at any instance */

  init_idec(&ne_orig);         /* create the i.dec file of global elmos*/

  rinput_init(filename_sav);

  if(choice)
    node_para_err(filename_sav);

  if((seg = calloc(NSGT*2,sizeof(int))) == NULL)
    {printf("out of memory for seg[]\n"); exit(1);}

  t1 = get_time() - tim;
  for(it = 0; it < nrec; it++)
  {
    if(fmod(it,2) == 0) flag = 0; else flag = 1;
    nsd = pow(2, it);     /* no of subdomains to be generated in this recursion*/

    filer(flag,choice,&fp,&fp2,&fp3,&fp4,&fp5,&fp6);

    for(rc = 0; rc < nsd; rc++)
    {
      rinput(fp);  /*read in the input.dat file for the current iter */
      rindec(it,fp3); /*read in the global elem no. info file for this iter */
      if(choice)
      {
	rinmen(fp5);   /* read file.men */
      }

      mfa(fpn, tmp, alpha); /* divide the mesh using MFA */

      partition(fp2,fp4,fp6); /*compile data structure for each partition
				 and write*/
    }
    printf("\nRecursion no.%3d completed\n\n", it+1);
    fclose(fp);
    fclose(fp2);
    fclose(fp3);
    fclose(fp4);
    fclose(fp5);
    fclose(fp6);
  }
  tim = get_time() - t1;
  fprintf(fpn,"\nTotal run-time =\t%.4f seconds", tim);
  fclose(fpn);
  printf("\nThe MFA results are in --------------------------> mfa.dat\n");

  free(mark1);
  free(mark2);

  if(choice)
  {
    free(hhi);
    free(cg);
    free(err);
    free(hi);
    free(seg);
  }

  if(fmod(nrec,2) == 0)
  {
    copy("i.dat","o.dat");
    copy("i.dec","o.dec");
    if(choice)
      copy("i.men","o.men");
  }

  form_decomp_dec(ne_orig);

  free(cord);
  free(nod);


  printf("The Decomposed File Generated as ----------------> o.dat\n");
  if(choice)
    printf("The Associated Mesh File is -------------------> o.men\n");

  strcpy(filename, filename_sav);
  strcat(filename,".dec");
  copy("decomp.dec",filename);
  printf("The Associated Plot File is ---------------------> %s\n",filename);
  printf("\n\n\tTotal time =>>>>>>> %.4f seconds ", tim); 

  remove("i.dat");
  remove("i.dec");
  remove("o.dec");
  remove("decomp.dec");

  return;

}


void  mfa(FILE *fpn, float tmp, float alpha)
{

float   energy, cost, penalty, sum_penalty, sum_penalty_tot;
int     numface;
float     tot_pred_elems;
int     i, j;

/*** allocate dynamic space for vectors v[] and s[] and matrix t[][] ***/

       if((t = calloc(ne*ne,sizeof(int))) == NULL)
	 {printf("out of memory for t[][]\n"); exit(1);}

       if((v = calloc(ne,sizeof(double))) == NULL)
	 {printf("out of memory for v[]\n"); exit(1);}

       if((s = calloc(ne,sizeof(float))) == NULL)
	 {printf("out of memory for s[]\n"); exit(1);}



/*** set up the weight matrix of connecting interfaces between elements ***/
/* the matrix consists of 0's for no connection and interface lengths */

weight_matrix();


/*** initialize v[] vector with random numbers ***/

set_v();


fprintf(fpn,"Temperature =\t%f ", tmp);
fprintf(fpn,"\nAlpha =\t%f ", alpha);


/* find the total predicted no of elemnts in the mesh */

    if(choice)
    {
      tot_pred_elems = 0.0;
      for(i = 0; i < ne; i++)
       tot_pred_elems += (float)err[i];
    }


/****** Perform Mean Field Annealing   ******/

anneal(tmp, alpha);


/* form the partiotioned results on the basis of vector v */

    for(i=0; i<ne; i++)
    {
      if (v[i] > 0.0)  s[i] = 1.0;
      else  s[i] = -1.0;
    }



/* compute the cost, penalty and the total energy */

    cost=0.0; sum_penalty=0.0; penalty=0.0; energy=0.0;
    for (i=0; i<ne; i++)
    {
       for (j=0; j<ne; j++)
       {
	 if(i !=j)
	 cost +=  (float)t[j+i*ne] * s[i] * s[j];
       }
    }
    cost *= -0.5;

    for (i=0; i<ne; i++)
    {

	 if(choice) sum_penalty += s[i] * (float)err[i];
	 else sum_penalty += s[i];
    }

    sum_penalty_tot *= sum_penalty;
    penalty = alpha * 0.5 * sum_penalty_tot;
    energy = cost + penalty;


/* calculate number of interfaces between partitions */

    numface=0;
    for(i=0; i<ne; i++)
	for(j=0; j<ne; j++)
	  if(s[i] != s[j] && t[j+i*ne] == 1)
	    numface++;

    numface = numface/2;


/* write mfa info to file mfa.out */

outfile_mfa(fpn, tot_pred_elems, sum_penalty, penalty, cost, numface, energy);


/* print results to the screen */

    if(choice)
    {
      printf("\nTotal no. of predicted elements  %f", tot_pred_elems);
      printf("\nSum penalty of elements  %f", sum_penalty);
    }

    printf("\n\nNumber of iterations for annealing -----> %d", mfa_it);
    printf("\nNumber of interfaces -----------------------> %d", numface);
    printf("\nPenalty --------------------------------->%f", penalty);
    printf("\nTotal energy -----------------------------> %f", energy);
    printf("\n");


/* put the partitions in different parts  */

sort_parts();


fprintf(fpn,"\nInput vector S for bisection 1:\n");
 for(i = 0; i < ne1; i++)  fprintf(fpn,"%.1f  ",s[i]);

fprintf(fpn,"\nInput vector S for bisection 2:\n");
 for(i = 0; i < ne2; i++)  fprintf(fpn,"%.1f  ",s[i]);

fprintf(fpn,"\n\n");

} /* end of mfa */




void  outfile_mfa(FILE *fpn, float tot_pred_elems, float sum_penalty,
	 float penalty, float cost, int numface, float energy)
{
int  i;


fprintf(fpn,"\nNumber of Iterations =\t%d", mfa_it);
if(choice)
{
  fprintf(fpn,"\nTotal no. of predicted elements =\t%f", tot_pred_elems);
  fprintf(fpn,"\nSum penalty of elements =\t%f", sum_penalty);
}
fprintf(fpn,"\nCost =\t%f ", cost);
fprintf(fpn,"\nNumber of faces =\t%d ", numface);
fprintf(fpn,"\nPenalty =\t%f", penalty);
fprintf(fpn,"\nEnergy =\t%f", energy);

}  /* end of infile_mfa()  */



void   weight_matrix()
{
int     a, b, i, j, k, m, n, counter, x[3],xx[3], nd[2];

      if(ne > 0)
    {
      for(i = 0; i < ne; i++)
      {
	for(j=0; j<3; j++)
	  x[j] = nod[4*i+j];
	for(k=0; k<ne; k++)
	{
	  counter=0;
	  if (k == i)  t[k+i*ne] = 0;
	  else
	  {
	    for(m=0; m<3; m++)
	      xx[m] = nod[4*k+m];
	    for(a=0; a<3; a++)
	    {
	      for(b=0; b<3; b++)
	      {
		if(x[a] == xx[b])
		{
		  nd[counter] = x[a];
		  counter++;
		}
		if(counter==2) break;
	      }
	      if(counter == 2)break;
	    }
	  }
	  if(counter == 2)
	    t[k+i*ne] = 1;

	  else t[k+i*ne] = 0;
	}
      }
    }

}  /*  end of weight_matrix()  */



void  set_v()
{
double  rd;
int     i;

   srand((unsigned) (long)seed);

    for (i=0; i<ne; i++)
    {
       rd=(float)(rand()%32768)/(double)(1L<<15);
       if(rd<0.5) v[i] = -1.0*0.0001*rd;
       else  v[i] = 1.0*0.0001*rd;
    }

}  /* end of random_v()  */



void  anneal(float tmp, float alpha)
{
double  sum, sat, sat1;
int     i, j;

    mfa_it = 1;

    while (1)
    {
      sat = 0.0;
      sat1=0.0;

      for (i=0; i<ne; i++)
      {
	 sum=0.0;
	 for (j=0; j<ne; j++)
	 {
	   if(j != i)
	   {  
	     if(choice) sum += ((double)t[j+i*ne] -
		    ((double)alpha * err[j])) * v[j] / (double)tmp;
	     else sum += ((double)t[j+i*ne] -
			 (double)alpha) * v[j] / (double)tmp;
	   } /* end of if statement */
	 } /* end of inner for loop */

	 sat1 = v[i];
	 v[i] = tanh(sum);          /* new value for V_i */
	 sat += sat1 * v[i];

      } /* end of outer for loop */

      sat /= (double)ne;   /* saturation */

      if(sat > 0.999 || mfa_it > 100) break;
      mfa_it++;

      tmp *= 0.95;         /* cool temperature by 0.95 */

      if(alpha < 1.0)  alpha *= 1.5;  /* inverse anneal alpha */
      else alpha = 1.0;


    } /* end of while loop */

}  /*  end of anneal()  */




void  sort_parts()
{
register i;
int eno;

	 mark1 = calloc(ne, sizeof(int));
	 if(mark1 == NULL){printf("out of core for mark1\n");exit(1);}

	 mark2 = calloc(ne, sizeof(int));
	 if(mark2 == NULL){printf("out of core for mark2\n");exit(1);}

      divider_1(s);/*divide ne into ne1 & ne2 on left and right of divider*/

     left_val = 0.; right_val = 0.;


      if(choice)
      {
	for(i = 0; i < ne1; i++)
	{
	  eno = mark1[i]-1;
	  left_val += err[eno];    /*the predicted no of elemnts*/
	}

	for(i = 0; i < ne2; i++)
	{
	  eno = mark2[i]-1;
	  right_val += err[eno];
	}
      }

      else
      {
	    left_val = (double) ne1;
	    right_val = (double) ne2;
      }

}





void divider_1()
{
register i;

/* allocate cg[2*ne], mark1[ne] and mark2[ne] arrays before hand.
   Elements left to divider are in mark1[] and on right in mark2[].        */
   for(i = 0; i < ne; mark1[i++] = 0);
   for(i = 0; i < ne; mark2[i++] = 0);

   ne1 = ne2 = 0;
   for(i = 0; i < ne; i++)
   {

      if(s[i] == -1.0)
      {
	mark1[ne1] = i+1;
	(ne1)++;
      }
      else
      {
	mark2[ne2] = i+1;
	(ne2)++;
      }

   }
}



void elemcord_del(int n,int nnn[], double x[], double y[], double del[])
{
register i;
   for(i = 0; i < 3; i++)
   {
   nnn[i] = nod[4*n+i];
   nodecord_del(nnn[i],&x[i], &y[i], &del[i]);
   }
}



void nodecord_del(int node, double *px, double *py, double *pdel)
{
   *px = cord[3*(node-1)+0];
   *py = cord[3*(node-1)+1];
   *pdel = hhi[node-1];
}


void seglen(double x1,double y1, double x2, double y2, double *pL)
{
*pL=sqrt(pow(x2-x1,2.)+pow(y2-y1,2.));
}


int interseg(double x1,double y1,double x2,double y2,
	     double x3,double y3,double x4,double y4, double *x, double *y)
{
/* [-1] non-intersecting segments
   [0]  parallel segments
   [1]  intersecting segments     */

double a11=x2-x1;
double a12=x3-x4;
double a21=y2-y1;
double a22=y3-y4;
double c11=x3-x1;
double c21=y3-y1;
double D=a11*a22-a12*a21;
double t,s;



   if(fabs(D) < 1.e-10)
   return(0);                    /*lines are parallel*/

   t=(a22*c11-a12*c21)/D;
   s=(a11*c21-a21*c11)/D;

   *x=x1+t*(x2-x1);
   *y=y1+t*(y2-y1);

   if(fabs(t) <= 1.e-10) t = 1.e-10;

   if(fabs(s) <= 1.e-10) s = 1.e-10;

   if(fabs(t-1.) <= 1.e-10) t = .998;

   if(fabs(s-1.) <= 1.e-10) s = .998;

   if(t < .999 && t > 0. && s < .999 && s > 0.)
   return(1);                     /* lines intersect */
   else
   return(-1);
}



double neuron(double x[], double y[], double del[] )
{
double L[3],A[3], tot_A, Yin[9], Yout[1], rms;

     calcang(x,y,L,A);
     tot_A = A[0]+A[1]+A[2];

     if(fabs(tot_A - 180.) > 1.e-1)
     {
     fprintf(stdout,"*** Record Suspicious Sum_Angles = %f***\n", tot_A);
     exit(1);
     }
     homogen_load(L,A,del,Yin);
     hoex15(Yin, Yout);

return(Yout[0]);
}


void calcang(double x[], double y[], double L[], double A[])
{
int i,j,k;
double Aa,Bb,D;
   j = 0;
   k = 0;
   for(i=0; i < 3; i++)
   {

   j = i+1;   if(j > 2) j=0;

   k = j+1;   if(k > 2) k = 0;

   seglen(x[i],y[i],x[j],y[j],&Aa);
   seglen(x[j],y[j],x[k],y[k],&Bb);
   dot(x[j],y[j],x[i],y[i],x[j],y[j],x[k],y[k], &D);
   A[i]=acos(D/(Aa*Bb))*57.29577951;
   L[i] = Aa;
   }
}



void dot(double x1, double y1, double x2, double y2,
    double x3, double y3, double x4, double y4, double *pD)
{
   *pD = (x2-x1)*(x4-x3)+(y2-y1)*(y4-y3);
}



/* Wed Feb 19 21:00:13 1992 (hoex15.c) Recall-Only Run-time for <heox15> */
/*    Control Strategy is: <dbd> */
int hoex15(double Yin[9], double Yout[1] )
{
    static double  Xout[17];  /* work arrays */
    long   ICmpT;  /* temp for comparisons */

    /* *** WARNING: Code generated assuming Recall = 0 *** */

    /* Read and scale input into network */
    Xout[2] = Yin[0] * (0.18411967) + (-1.0475949);
    Xout[3] = Yin[1] * (0.13056705) + (-1.0534933);
    Xout[4] = Yin[2] * (0.16848206) + (-1.0864313);
    Xout[5] = Yin[3] * (0.032501669) + (-2.0954135);
    Xout[6] = Yin[4] * (0.043108462) + (-2.3042163);
    Xout[7] = Yin[5] * (0.027944278) + (-1.8454345);
    Xout[8] = Yin[6] + (-2);
    Xout[9] = Yin[7] * (0.27430327) + (-1.0311883);
    Xout[10] = Yin[8] * (0.34883314) + (-1.0618132);
LAB110:

    /* Generating code for PE 0 in layer 3 */
    Xout[11]  = 0.15789863 + 0.10383216 * Xout[2] + (-0.22523752) * Xout[3] +
	(-0.67347956) * Xout[4] + (-0.05143927) * Xout[5] +
	(-0.3063916) * Xout[6] + 0.14410186 * Xout[7] +
	(-0.053388517) * Xout[8] + 0.49070397 * Xout[9] + 0.05235222 * Xout[10]
	;
    Xout[11] = tanh( Xout[11] );

    /* Generating code for PE 1 in layer 3 */
    Xout[12]  = (-7.2727799) + 1.0029693 * Xout[2] + 1.1253526 * Xout[3] +
	(-0.096689932) * Xout[4] + 0.15600094 * Xout[5] + 0.10614354 * Xout[6]
	 + 0.061088312 * Xout[7] + (-0.035957791) * Xout[8] +
	(-6.1995211) * Xout[9] + (-2.2345977) * Xout[10];
    Xout[12] = tanh( Xout[12] );

    /* Generating code for PE 2 in layer 3 */
    Xout[13]  = (-1.8205116) + (-0.38261673) * Xout[2] + 1.2245123 * Xout[3] +
	1.0399791 * Xout[4] + 0.10367764 * Xout[5] + 0.22796212 * Xout[6] +
	0.14192601 * Xout[7] + (-0.079842076) * Xout[8] +
	(-1.2307496) * Xout[9] + (-1.8718977) * Xout[10];
    Xout[13] = tanh( Xout[13] );

    /* Generating code for PE 3 in layer 3 */
    Xout[14]  = (-0.93488657) + (-0.95023084) * Xout[2] +
	(-0.054171421) * Xout[3] + (-0.3570905) * Xout[4] +
	0.046458505 * Xout[5] + 0.1223551 * Xout[6] + (-0.21982495) * Xout[7]
	 + 0.065754555 * Xout[8] + (-0.10536617) * Xout[9] +
	(-0.051805291) * Xout[10];
    Xout[14] = tanh( Xout[14] );

    /* Generating code for PE 4 in layer 3 */
    Xout[15]  = 0.010340533 + (-0.13679253) * Xout[2] + 0.19975138 * Xout[3] +
	0.16966611 * Xout[4] + 0.099696361 * Xout[5] + 0.13113278 * Xout[6] +
	0.02222888 * Xout[7] + 0.061033741 * Xout[8] + (-0.047290441) * Xout[9]
	 + (-0.058939323) * Xout[10];
    Xout[15] = tanh( Xout[15] );

    /* Generating code for PE 0 in layer 4 */
    Xout[16]  = 0.11256779 + (-0.13167156) * Xout[11] + 0.77680171 * Xout[12]
	 + 0.36559567 * Xout[13] + (-0.37231231) * Xout[14] +
	0.033708036 * Xout[15];
    Xout[16] = tanh( Xout[16] );

    /* De-scale and write output from network */
    Yout[0] = Xout[16] * (63.749999) + (52);
    return( 0 );
}



void homogen_load(double *L, double *A, double *del, double *Yin)
{
double div;

div = del[0];
L[0]=L[0]/div;
L[1]=L[1]/div;
L[2]=L[2]/div;
del[0]=del[0]/div;
del[1]=del[1]/div;
del[2]=del[2]/div;
memmove(&Yin[0],&L[0], 3*sizeof(double));
memmove(&Yin[3],&A[0], 3*sizeof(double));
memmove(&Yin[6],&del[0], 3*sizeof(double));
}



/* function to read in the data from the primary data file */

void rinput(FILE *fp11)
{
char bf[256];
int i,j,k,n;
double x1,y1,z1;

  fgets(titel,50,fp11);
  fputs(titel,stdout);
  fgets(bf,256,fp11);
  sscanf(bf,"%lf%lf",&dt,&damp);
  printf("%f %f \n",dt,damp);
  fgets(bf,256,fp11);
  sscanf(bf,"%d %d %d %d %d %d %d %d %d %d %d",&nn,&ne,&nb,&nm,&nl,
  &ndime, &nstr, &nbe, &kmax,&nmax, &iln);

  fgets(bf,256,fp11);
  sscanf(bf,"%d",&nq);


  for(i = 0; i < nn; i++)
  {
    fgets(bf,256,fp11);
    sscanf(bf,"%d %lf %lf %lf",&n, &x1, &y1,&z1);
    cord[3*i] = x1;
    cord[3*i + 1] = y1;
    cord[3*i + 2] = z1;
  }

  if(ne > 0)
  {
      for(i = 0; i < ne; i++)
      {
	fgets(bf,256,fp11);
	sscanf(bf,"%d %d %d %d %d",&n,&nod[4*i + 0],&nod[4*i + 1],
	&nod[4*i + 2],&nod[4*i + 3]);
      }
   }

   if(nl > 0)
   {
      for(i = 0; i < nl;i++)
      {
	fgets(bf,256,fp11);
	sscanf(bf,"%d %d %d %d",&n,&nodl[i][0],&nodl[i][1],
	&nodl[i][2]);
      }
   }

   if(nq > 0)
   {
      for(i = 0; i < nq; i++)
      {
	fgets(bf,256,fp11);
	sscanf(bf,"%d %d %d %d %d %d",&n,&nodq[i][0],&nodq[i][1],
	&nodq[i][2],&nodq[i][3],&nodq[i][4]);
      }
   }

   for(i = 0; i < nm; i++)
   {
     fgets(bf,256,fp11);
     sscanf(bf,"%d %d %lf %lf %lf %lf %lf",&n,&mat[i],&amat[5*i+0],
     &amat[5*i+1],&amat[5*i+2],&amat[5*i+3],&amat[5*i+4]);
   }

   for(i = 0; i < nb; i++)
   {
     fgets(bf,256,fp11);
     sscanf(bf,"%d %d %d %d",&nco[i][0],&nco[i][1],&nco[i][2],&nco[i][3]);
   }



   if(iln > 0)
   {
     for(i = 0; i < NC; nc[i++] = 0);/*zero for use in nb_iln */
	iln = 0;

     more:
     iln++;
     i = iln - 1;

     fgets(bf,256,fp11);
     sscanf(bf,"%d %lf %lf %lf",&nc[i],&aload[3*i+0],
     &aload[3*i+1],&aload[3*i+2]);

     if(nc[i] < nn)
       goto more;
   }

   if(nstr > 0)
   {
     for(i = 0; i < nstr; i++)
     {
       fgets(bf,256,fp11);
       sscanf(bf,"%d",&ngs[i]);
       fgets(bf,256,fp11);
       sscanf(bf,"%d %d %d %d %d %d %d %d",&nsnod[i][0],&nsnod[i][1],
       &nsnod[i][2],&nsnod[i][3],&nsnod[i][4],&nsnod[i][5],&nsnod[i][6],
       &nsnod[i][7]);
     }
   }
}


void rindec(int it, FILE *fp3)
{
register i;
int dum;
char buf[100];

  switch(it)
  {
    case 0:
    if((ne_g = calloc(ne, sizeof(int))) == NULL)
    {
      printf("out of core for ne_g\n"); exit(1);
    }
    break;

    default:
    break;
  }

  fgets(buf,100,fp3);

  for(i = 0; i < ne; i++)
  {
    fgets(buf,100,fp3);
    sscanf(buf,"%d %d",&dum, &ne_g[i]);
  }

}



void rinmen(FILE *fp5)
{
register i;
int dum;
char buf[100];

  fgets(buf,100,fp5);       /* read header line */
  for(i = 0; i < nn; i++)
  {
    fgets(buf,100,fp5);
    sscanf(buf,"%d %lf",&dum, &hhi[i]);
  }

  fgets(buf,100,fp5);       /* read header line */
  for(i = 0; i < ne; i++)
  {
    fgets(buf,100,fp5);
    sscanf(buf,"%d %lf",&dum, &err[i]);
  }

  fgets(buf,100,fp5);       /* read header line */

  for(i = 0; i < ne; i++)
  {
    fgets(buf,100,fp5);
    sscanf(buf,"%d %lf",&dum, &hi[i]);
  }


}


/* function to write in the data to one of the two subdomain files*/

void routput(FILE *fp00)
{
int i,j,k,n, n1,n2,n3,n4;

  fputs((SUB.titel),fp00);
  fprintf(fp00,"%f %f\n",SUB.dt,SUB.damp);
  fprintf(fp00,"%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d\n",
  SUB.nn,SUB.ne,SUB.nb,SUB.nm,SUB.nl,
  SUB.ndime, SUB.nstr, SUB.nbe, SUB.kmax,SUB.nmax, SUB.iln);


  fprintf(fp00,"%5d\n",SUB.nq);

  for(i = 0; i < SUB.nn; i++)
  {
     fprintf(fp00,"%5d%15.5f%15.5f%15.5f\n",i+1,
     (SUB.cord)[3*i + 0], (SUB.cord)[3*i + 1], (SUB.cord)[3*i + 2]);
  }

  if(SUB.ne > 0)
  {
    for(i = 0; i < SUB.ne; i++)
    {
      n1 = SUB.nod[4*i + 0];
      n2 = SUB.nod[4*i + 1];
      n3 = SUB.nod[4*i + 2];
      n4 = SUB.nod[4*i + 3];

      fprintf(fp00,"%5d%5d%5d%5d%5d\n",i+1,n1,n2,n3,n4);
    }
  }



  if(SUB.nl > 0)
  {
     for(i = 0; i < SUB.nl;i++)
     {
	fprintf(fp00,"%5d%5d%5d%5d\n",i+1,
	SUB.nodl[i][0],SUB.nodl[i][1],SUB.nodl[i][2]);
     }
  }


  for(i = 0; i < SUB.nm; i++)
  {
    fprintf(fp00,"%5d%5d%12.5e%12.5e%12.5e%12.5e%12.5e\n",i+1,
    (SUB.mat)[i],(SUB.amat)[5*i+0],
    (SUB.amat)[5*i+1],(SUB.amat)[5*i+2],(SUB.amat)[5*i+3],(SUB.amat)[5*i+4]);
  }

  for(i = 0; i < SUB.nb; i++)
  {
    fprintf(fp00,"%5d%5d%5d%5d\n",(SUB.nco)[i][0],
    (SUB.nco)[i][1],(SUB.nco)[i][2],(SUB.nco)[i][3]);
  }

  if(SUB.iln > 0)
  {
    SUB.iln = 0;
    more:
    SUB.iln++;
    i = SUB.iln - 1;

    fprintf(fp00,"%5d%15.5f%15.5f%15.5f\n",
    (SUB.nc)[i],(SUB.aload)[3*i+0],(SUB.aload)[3*i+1],
    (SUB.aload)[3*i+2]);

    if(SUB.nc[i] < SUB.nn)
      goto more;
  }

  if(SUB.nstr > 0)
  {
    for(i = 0; i < SUB.nstr; i++)
    {
      fprintf(fp00,"%5d\n",SUB.ngs[i]);
      fprintf(fp00,"%5d%5d%5d%5d%5d%5d%5d%5d\n",
      SUB.nsnod[i][0],SUB.nsnod[i][1],
      SUB.nsnod[i][2],SUB.nsnod[i][3],
      SUB.nsnod[i][4],SUB.nsnod[i][5],
      SUB.nsnod[i][6],SUB.nsnod[i][7]);
    }
  }
}



void routdec(int nex, FILE *fp4)
{
register i;

  fprintf(fp4,"%5d          %5d\n",t_sds,nex);

  for(i = 0; i < nex; i++)
     fprintf(fp4,"%5d %5d\n",i+1, SUB.ne_g[i]);
}



void routmen(int nnx, int nex, FILE *fp6)
{
register i;

  fprintf(fp6,"node paras hhi[]\n");
  for(i = 0; i < nnx; i++)
     fprintf(fp6,"%5d   %lf\n",i+1, SUB.hhi[i]);
  fprintf(fp6,"eta_i err[]\n");
  for(i = 0; i < nex; i++)
     fprintf(fp6,"%5d   %lf\n",i+1, SUB.err[i]);
  fprintf(fp6,"elem sizes hi[]\n");
  for(i = 0; i < nex; i++)
     fprintf(fp6,"%5d   %lf\n",i+1, SUB.hi[i]);
}


/* The function loads in the data structure for a subdomain in SUB
each element identified by mark1[] & mark2[] has its global element
number passed on to this function                                */

void stack(int elem_g)
{
int i,ii,jj,test1;
int trace[3],node, mat_no;
double x[3], y[3], z[3], hhf[3];

  ii = SUB.nn ;
  jj = SUB.ne ;

  for(i = 0; i < 3; i++)
  {
    node = nod[4*(elem_g - 1) + i] - 1;/*load the coordinates of the element*/
    x[i] = cord[3*node + 0];
    y[i] = cord[3*node + 1];
    z[i] = cord[3*node + 2];
    hhf[i] = hhi[node];  
  }

  mat_no = nod[4*(elem_g - 1) + 3];/* load the mat number */

  for(i = 0; i < 3; i++)
  {

     test1 = 0;
     test1 = comp_st(x[i],y[i],z[i]); /*test1 returns the compared node_st*/

     if(test1 == -1)
     {
	 SUB.cord[3*ii + 0] = x[i];
	 SUB.cord[3*ii + 1] = y[i];
	 SUB.cord[3*ii + 2] = z[i];
	 SUB.hhi[ii]        = hhf[i]; 
	 trace[i] = ii + 1;
	 ii++;
	 continue;
     }
     trace[i] = test1;
   }

   /* load in the new topology */

   revise_nod(trace, &jj, mat_no);


   SUB.nn = ii;
   SUB.ne = jj;

   return;
}


/* function to compare the packet nodes with the stack nodes */

int comp_st(double x, double y, double z)
{
int i;
double x_st, y_st, z_st;
double tol = 1.e-4;

  for(i = 0; i < SUB.nn; i++)
  {
    x_st = SUB.cord[3*i + 0];
    y_st = SUB.cord[3*i + 1];
    z_st = SUB.cord[3*i + 2];
    if((fabs(x - x_st) <= tol) && (fabs(y - y_st) <= tol)
	  && (fabs(z - z_st) <= tol))
    return(i + 1);
  }
  return(-1);
}


/* function to revise the node topology */

void revise_nod(int trace[], int *pjj, int mat_no)
{
int jj = *pjj;

  SUB.nod[4*jj + 0] = trace[0];
  SUB.nod[4*jj + 1] = trace[1];
  SUB.nod[4*jj + 2] = trace[2];
  SUB.nod[4*jj + 3] = mat_no;
  jj++;

  *pjj = jj;

}


/* function to identify the restrained and loaded nodes from the
   compiled mesh */

void nb_iln(void)
{
int i,j,k,l,m,o, node;
double tol = pow(10.,-6);
double cx_b,cy_b,cz_b,cx_l,cy_l,cz_l;

  k = 0;
  m = 0;
  SUB.nb = 0;
  for(i = 0; i < SUB.nn; i++)
  {
    for(j = 0; j  < nb; j++)
    {
       node = nco[j][0] - 1;
       cx_b = fabs(SUB.cord[3*i + 0] - cord[3*node + 0]);
       cy_b = fabs(SUB.cord[3*i + 1] - cord[3*node + 1]);
       cz_b = fabs(SUB.cord[3*i + 2] - cord[3*node + 2]);
       if(cx_b <= tol && cy_b<= tol && cz_b <= tol)
       {
	  SUB.nco[k][0] = i + 1;
	  SUB.nco[k][1] = nco[j][1];
	  SUB.nco[k][2] = nco[j][2];
	  SUB.nco[k][3] = nco[j][3];
	  k++;
	  break;
       }
    }

    for(l = 0; nc[l] <= nn && nc[l] != LASTNODE; l++)
    {
	node = nc[l] - 1;
	cx_l = fabs(SUB.cord[3*i + 0] - cord[3*node + 0]);
	cy_l = fabs(SUB.cord[3*i + 1] - cord[3*node + 1]);
	cz_l = fabs(SUB.cord[3*i + 2] - cord[3*node + 2]);
	if(cx_l <= tol &&  cy_l <= tol && cz_l <= tol)
	{
	   SUB.nc[m] = i + 1;
	   for(o = 0; o < 3; o++)
		 SUB.aload[3*m+o] = aload[3*l+o];
	   m++;
	   break;
	}
    }
  }
  SUB.nb = k;

  if(m == 0)
  {
    SUB.nc[m] = LASTNODE;
    SUB.aload[3*m+0] = SUB.aload[3*m+1] = SUB.aload[3*m+2] = 0.;
    m++;
    SUB.iln = m;
  }
  else
  {
    SUB.nc[m] = LASTNODE;
    SUB.aload[3*m+0] = SUB.aload[3*m+1] = SUB.aload[3*m+2] = 0.;
    m++;
    SUB.iln = m;
  }
}




void partition(FILE *fp2, FILE *fp4, FILE *fp6)
{
register i,l;
int elem_g;

  memmove((SUB.titel),titel, sizeof(titel));
  memmove((SUB.mat), mat, sizeof(mat));
  memmove((SUB.amat),amat,sizeof(amat));
  SUB.nn = 0;
  SUB.ne = 0;
  SUB.nq = 0;
  SUB.dt = dt;
  SUB.damp = damp;
  SUB.ndime = ndime;
  SUB.nm =  nm;
  SUB.nl = nl;
  SUB.nstr = nstr;
  SUB.nbe = nbe;
  SUB.kmax = kmax;
  SUB.nmax = nmax;


  for(i = 0; i < ne1; i++)
  {
    elem_g = mark1[i];
    stack(elem_g);
    SUB.ne_g[i] = ne_g[elem_g-1];
    if(choice)
    {
      SUB.err[i]  =  err[elem_g-1];
      SUB.hi[i]   =   hi[elem_g-1];
    }
  }
  nb_iln();
  routput(fp2);
  routdec(ne1,fp4);
  if(choice)
    routmen(SUB.nn,ne1,fp6);

  SUB.nn = 0;
  SUB.ne = 0;
  for(i = 0; i < ne2; i++)
  {
    elem_g = mark2[i];
    stack(elem_g);
    SUB.ne_g[i] = ne_g[elem_g-1];
    if(choice)
    {
       SUB.err[i]  =  err[elem_g-1];
       SUB.hi[i]   =   hi[elem_g-1];
    }
  }
  nb_iln();
  routput(fp2);
  routdec(ne2,fp4);
  if(choice)
     routmen(SUB.nn,ne2,fp6);

}


void copy(char *arg1, char *arg2)
{
FILE *fp1,*fp2;
char *buf;
long f_size;

  if((fp1=fopen(arg1,"rb")) == (FILE *) NULL)
    {printf("cannot open %s\n",arg1); exit(1);}

  if((fp2=fopen(arg2,"wb")) == (FILE *) NULL)
    {printf("cannot open %s\n",arg2); exit(1);}

  fseek(fp1,0,SEEK_END);
  f_size = ftell(fp1);
  rewind(fp1);
  if((buf=malloc(f_size)) == NULL)
    {printf("out of memory\n");exit(1);}
  fread(buf,f_size,1,fp1);
  fwrite(buf,f_size,1,fp2);
  free(buf);
  fclose(fp1);
  fclose(fp2);
}



void init_idec(int *ne_orig)
{
register i;
FILE *fl1, *fl2;
char buf[256];

/* creates the first i.dec file containing the global element numbers
for the initial domain and  allocates men[] if choice == [1]*/

  if((fl1 = fopen("i.dat","r")) == (FILE *) NULL)
    {printf("cannot open i.dat in init_dec()\n"); exit(1);}
  fgets(buf,256,fl1);
  fgets(buf,256,fl1);
  fgets(buf,256,fl1);
  sscanf(buf,"%d %d %d %d %d %d %d %d %d %d %d",&nn,&ne,&nb,&nm,&nl,
  &ndime, &nstr, &nbe, &kmax,&nmax, &iln);
  *ne_orig = ne;
  printf("ne = %5d\n",ne);
  fclose(fl1);

  if((fl2 = fopen("i.dec","w")) == (FILE *) NULL)
    {printf("cannot open i.dec\n"); exit(1);}
  fprintf(fl2,"%5d            %5d\n",1,ne);
  for(i = 0; i < ne; i++)
     fprintf(fl2,"%5d %5d\n",i+1, i+1);
  fclose(fl2);
}



void form_decomp_dec(int ne_orig)
{
FILE *ff,*ff1;
char bf[256];
register i,j;
int dum, ne_sd;

  ne_sd = ne_orig/t_sds + UnBalance; /* get an estimate on the no of
					element in each subdomain */

  if((elems_sd = calloc(ne_sd,sizeof(int))) == NULL)
    {printf("cannot allocate for elems_sd in form_decopm_dec\n");exit(1);}
  if((elem_sdno = calloc(ne_orig, sizeof(int))) == NULL)
    {printf("cannot allocate for elem_sdno in form_decopm_dec\n");exit(1);}

  if((ff= fopen("o.dec","r")) == (FILE *) NULL)
   {printf("cannot open o.dec for reading in form_decomp_dec()\n");exit(1);}
  if((ff1= fopen("decomp.dec","w")) == (FILE *) NULL)
    {printf("cannot open decomp.dec for reading in form_decomp_dec()\n");exit(1);}

  for(i = 0;  i < t_sds; i++)
  {
     fgets(bf,256,ff); sscanf(bf,"%d %d", &dum, &ne_sd);

     for(j = 0; j < ne_sd; j++)
     {
       fgets(bf,256,ff);
       sscanf(bf,"%d %d", &dum,&(elems_sd[j]));
     }

     load_elem_sdno(i, ne_sd);

  }
  free(elems_sd);
  fclose(ff);

  for(i = 0; i < ne_orig; i++)
     fprintf(ff1,"%f\n",(float) elem_sdno[i]);

  fclose(ff1);
  free(elem_sdno);

}



void load_elem_sdno(register i, int ne_sd)
{
  register j, pos;

  for(j = 0;j < ne_sd; j++)
  {
    pos = elems_sd[j]-1;
    elem_sdno[pos] = i+1;
  }

}



void node_para_err(char *file_sav)
{
char bf[70];
FILE *fme,*fmen;
int i,j,k, nnn[3];
double xx[3], yy[3], del[3];

/* function generates the node paras and reads elem errors.
   A file with extension .men is generated which has a the
   following format :-
   header line ******************
   1   node para_1
   .       .
   nn  node para_nn
   header line ******************
   1  predicted no of elemnts err_1
   .                            .
   ne                         err_ne
   header line  *****************
   1  elem size hi_1
   .             .
   ne           hi_ne
			       */



  if((sort=calloc(9*nn,sizeof(int)))==NULL)
    {printf("out of memory for sort[]\n");exit(1);}



  search_node();

  if((hi=calloc(ne,sizeof(double)))==NULL)
    {printf("out of memory for hi[]\n");exit(1);}

  if((hhi=calloc(nn,sizeof(double)))==NULL)
    {printf("out of memory for hhi[]\n");exit(1);}

  if((err=calloc(ne,sizeof(double)))==NULL)
    {printf("out of memory for ne\n");exit(1);}

  strcpy(bf, file_sav);strcat(bf,".me");
  if((fme=fopen(bf,"r"))==(FILE *) NULL)
    {printf("cannot open %s\n", bf);exit(1);}

  for(i = 0; i < ne; i++)      /* read in *.me */
  {
    fgets(bf,70,fme);
    sscanf(bf,"%d %lf", &j, &hi[i]);
  }
  fclose(fme);

/* avg the hi values w.r.t. the connecting elements */
  for(i = 0; i < nn; i++)
  {
     for(j = 0; (k = sort[7*i+j]) != 0 ; j++)
	hhi[i] += hi[k-1];

     if(j == 0)
	continue;

     hhi[i] = hhi[i]/(double) j;

   }

   free(sort);


   strcpy(bf, file_sav);strcat(bf,".men");
   if((fmen=fopen(bf,"w"))==(FILE *) NULL)
     {printf("cannot open %s\n", bf);exit(1);}

   fprintf(fmen,"node paras hhi[]\n");        /* header line */
   for(i = 0; i < nn; i++)
      fprintf(fmen,"%3d    %lf\n",i+1,hhi[i]);



   fprintf(fmen,"predicted nelem err[]\n");



   for(i = 0; i < ne; i++)
   {
      elemcord_del(i, nnn, xx, yy, del);
      err[i] = neuron(xx, yy, del);
      if(err[i] <= 0.) err[i] = 1.;
      fprintf(fmen,"%3d    %lf\n",i+1, err[i]);
   }

   fprintf(fmen,"elem sizes hi[]\n");

   for(i = 0; i < ne; i++)
   {
      fprintf(fmen,"%3d    %lf\n",i+1, hi[i]);
   }

   fclose(fmen);

   strcpy(bf, file_sav); strcat(bf,".men");

   copy(bf,"i.men");
}



void search_node(void)
{

int i,j,k,l;

   for(i=0; i < nn; i++)
      for(j=0; j < 7; j++)
	sort[7*i+j]= 0;

/* Begin searching and sorting the elements connecting to a node,
at a time */


   for(i=0; i < nn; i++)
   {
     l=0;
     for(j=0; j < ne; j++)
     {
	  for(k=0; k < 3; k++)
	  {
	     if((i+1) == nod[4*j+k])
	     {
	       sort[7*i+l]= j+1;
	       l++;
	       break;
	     }
	  }
      }
   }

   return;
}



void rinput_init(char *file)
{
FILE *finp;
char bf[256];
int i,j,k,n;
double x1,y1,z1;

  strcpy(bf,file);
  strcat(bf,".dat");
  if((finp=fopen(bf,"r"))==(FILE *)NULL)
    {printf("cannot open %s\n",bf); exit(1);}

  fgets(bf,50,finp);
  fgets(bf,256,finp);
  fgets(bf,256,finp);
  sscanf(bf,"%d %d %d %d %d %d %d %d %d %d %d",&nn,&ne,&nb,&nm,&nl,
   &ndime, &nstr, &nbe, &kmax,&nmax, &iln);

  fgets(bf,256,finp);
  sscanf(bf,"%d",&nq);

  if((cord = calloc(3*nn, sizeof(double))) == NULL)
    {printf("out of core for cord\n"); exit(1);}

  if((nod = calloc(4*ne, sizeof(int))) == NULL)
    {printf("out of core for nod\n"); exit(1);}

  for(i = 0; i < nn; i++)
  {
    fgets(bf,256,finp);
    sscanf(bf,"%d %lf %lf %lf",&n, &x1, &y1,&z1);
    cord[3*i] = x1;
    cord[3*i + 1] = y1;
    cord[3*i + 2] = z1;
  }



  if(ne > 0)
  {
      for(i = 0; i < ne; i++)
      {
	fgets(bf,256,finp);
	sscanf(bf,"%d %d %d %d %d",&n,&nod[4*i + 0],&nod[4*i + 1],
	 &nod[4*i + 2],&nod[4*i + 3]);
      }
  }

  fclose(finp);
}



void form_boun(void)
{
register i;

  bsgt = 0; /* global variable */


  for(i = 0; i < ne; i++)
     compare_and_stack(i);

}



void compare_and_stack(int i)
{
register j,k,l,m,n;
int  nd1, nd2, nf1, nf2, knt;

  for(k = 0; k < 3; k++)
  {
     l = k+1; if(l>2) l = 0;
     nd1 = nod[4*i+k];  nd2 = nod[4*i+l];

     knt = 0;
     for(j = 0; j < ne; j++)
     {
       if(i == j) continue;

       for(m = 0; m < 3; m++)
       {
	  n = m+1; if(n>2) n = 0;
	  nf1 = nod[4*j+m]; nf2 = nod[4*j+n];
	  if((nf1 == nd1 && nf2 == nd2) || (nf2 == nd1 && nf1 == nd2))
	  {
	    knt++;
	    break;
	  }
       }

     }
     if(knt == 0)
     stack_seg(nd1, nd2);
  }

}



void stack_seg(int nd1, int nd2)
{
register i,j;

  for(i = 0; i < bsgt; i++)
  {
  if((nd1 == seg[2*i+0] && nd2 == seg[2*i+1]) ||
    (nd2 == seg[2*i+0] && nd1 == seg[2*i+1]))
       return;
  }
  seg[2*bsgt+0] = nd1;
  seg[2*bsgt+1] = nd2;
  bsgt++;
}



void filer(int flag, int choice, FILE **fp, FILE **fp2, FILE **fp3, FILE **fp4,
	   FILE **fp5, FILE **fp6)
{
  if(flag == 0)
  {
    if((*fp = fopen("i.dat","r")) == (FILE *) NULL)
      {printf("cannot open i.dat(r)\n"); exit(1);}
    if((*fp2 = fopen("o.dat","w")) == (FILE *) NULL)
      {printf("cannot open o.dat(w)\n"); exit(1);}

    if((*fp3 = fopen("i.dec","r")) == (FILE *) NULL)
      {printf("cannot open i.dec(r)\n"); exit(1);}
    if((*fp4 = fopen("o.dec","w")) == (FILE *) NULL)
      {printf("cannot open o.dec(w)\n"); exit(1);}

    if(choice)
    {
      if((*fp5 = fopen("i.men","r")) == (FILE *) NULL)
	{printf("cannot open i.men(r)\n"); exit(1);}
      if((*fp6 = fopen("o.men","w")) == (FILE *) NULL)
	{printf("cannot open o.men(w)\n"); exit(1);}
    }
  }
  else
  {
    if((*fp = fopen("o.dat","r")) == (FILE *) NULL)
      {printf("cannot open o.dat(r)\n"); exit(1);}
    if((*fp2 = fopen("i.dat","w")) == (FILE *) NULL)
      {printf("cannot open i.dat(w)\n"); exit(1);}

    if((*fp3 = fopen("o.dec","r")) == (FILE *) NULL)
      {printf("cannot open o.dec(r)\n"); exit(1);}
    if((*fp4 = fopen("i.dec","w")) == (FILE *) NULL)
      {printf("cannot open i.dec(w)\n"); exit(1);}

    if(choice)
    {
       if((*fp5 = fopen("o.men","r")) == (FILE *) NULL)
	 {printf("cannot open i.men(r)\n"); exit(1);}
       if((*fp6 = fopen("i.men","w")) == (FILE *) NULL)
	 {printf("cannot open o.men(w)\n"); exit(1);}
    }
  }

}



void  IO_print(target)
int    target;
{
int i, j, n;

   n = strlen(IO_str);
   for (i = j = 0; i < n; i++)
   {
      IO_wkstr[j++] = IO_str[i];
      if ((i<(n-1))&&(IO_str[i]=='\n'))
      IO_wkstr[j++] = ' ';
   } /* end for i */
   IO_wkstr[j] = '\0';
   strcpy (IO_str, IO_wkstr);
   target*=1; /* avoids warnings which occur when 'target' is unused */
printf(" %s", IO_str);
   
} /* IO_print */



float  IO_my_get_float()
{
   float  num;

   while (1)
   {
      if ( (gets(in_string) != NULL)
	   && (sscanf(in_string, "%f", &num) == 1) )
	 break;
      else 
      {
	 sprintf(IO_str, "\n   sorry, I don't understand.  Try again: ");
	 IO_print(0);
      }
   }
   return(num);

}  /* IO_my_get_float */



int  IO_my_get_int()
{
   int  num;

   while (1)
   {
      if ( (gets(in_string) != NULL)
	   && (sscanf(in_string, "%d", &num) == 1) )
	 break;
      else
      {
	 sprintf(IO_str, "\n   sorry, I don't understand.  Try again: ");
	 IO_print(0);
      }
   }
   return(num);

} /* IO_my_get_int */



float get_time()
{
  return((float)clock()/CLK_TCK);

} /* time function */

