/*                      Worker code for Potts

   USES MEAN FIELD ANNEALING POTTS CODE FOR SUB-DOMAIN GENERATION
   This program uses the Mean Field Anealing Potts NN to optimize
   the partition of a FE mesh.
   If choice is 0 then mesh is partitioned without the predictive
   model, if choice is 1 the mesh is partitioned using the predictive
   mode  (See the following paper)

   Subdomain Generation Using Parallel Q-State Potts Neural Networks
   A. Bahreininejad, B. H. V. Topping

   This version of software can be compiled along with the given
   configuration files for a pipeline of transputers.
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <dos.h>
#include  <chan.h>
#include  <stddef.h>


CHAN    **in_p, **out_p;
int     num_proc;
int     proc_id;

int     ne, *nod, choice;
double  *err;

double   *v;
int      *t;          /* vector s for the partitioned result */


void    mfa_potts_worker(double tmp, double alpha);
void    weight_matrix();
void    anneal_worker(double tmp, double alpha);
void    rec_v();

main(argc, argv, envp, in_ports, ins, out_ports, outs)

int     argc, ins, outs;
char    *argv[], *envp[];
CHAN    *in_ports[], *out_ports[];

{
double   tmp, alpha; /* tmp is the temperature and alpha is imbalance*/
int      i, j, k,msl_tmp, msl_alpha, msl_nod, msl_err;
double   p_elems;

    in_p = in_ports;
    out_p = out_ports;

    msl_tmp = sizeof(double);
    msl_alpha = sizeof(double);

    /* receive worker id's and number of workers */

    chan_in_word(&proc_id, in_p[1]);
    proc_id++; 
    chan_in_word(&num_proc, in_p[1]);

    if(proc_id != num_proc)
    {
       chan_out_word(proc_id, out_p[0]);
       chan_out_word(num_proc, out_p[0]);
    }

    /* receive the info of what type of decomposition */

    if(proc_id == num_proc)
      chan_in_word(&choice, in_p[1]);
    else
    {
	    chan_in_word(&choice, in_p[1]);
	    chan_out_word(choice, out_p[0]);
    }



    /* receive temperature and alpha */

    if(proc_id == num_proc)
    {
	  chan_in_message(msl_tmp, &tmp, in_p[1]);
	  chan_in_message(msl_alpha, &alpha, in_p[1]);
	  chan_in_word(&ne, in_p[1]);
    }
    else
    {
	  chan_in_message(msl_tmp, &tmp, in_p[1]);
	  chan_in_message(msl_alpha, &alpha, in_p[1]);
	  chan_in_word(&ne, in_p[1]);
	  chan_out_message(msl_tmp, &tmp, out_p[0]);
	  chan_out_message(msl_alpha, &alpha, out_p[0]);
	  chan_out_word(ne, out_p[0]);
    }



    /* receive the nodes and the number of predicted elements */

    nod = calloc(4*ne, sizeof(int));
    msl_nod = 4*ne*sizeof(int);
    if(proc_id == num_proc)   
      chan_in_message(msl_nod, nod, in_p[1]);
    else
    {
	   chan_in_message(msl_nod, nod, in_p[1]);              
	   chan_out_message(msl_nod, nod, out_p[0]);    
    }

     
     if(choice)
     {
	err=calloc(ne,sizeof(double));
	msl_err = sizeof(double);
	  if(proc_id == num_proc)
	  {
	     for(j=0; j<ne; j++)
	     {
		  chan_in_message(msl_err, &p_elems, in_p[1]);
		  err[j] = p_elems;
	     }
	  }
	  else
	  {
	     for(i=0; i<num_proc-proc_id; i++)
	     {
		 for(j=0; j<ne; j++)
		 {
		    chan_in_message(msl_err, &p_elems, in_p[1]);
		    err[j] = p_elems;
		    chan_out_message(msl_err, &p_elems, out_p[0]);
		  }
	     }
	     for(j=0; j<ne; j++)
	     {
	       chan_in_message(msl_err, &p_elems, in_p[1]);
	       err[j] = p_elems;
	     }
	  }
     }


  mfa_potts_worker(tmp, alpha); /* divide elements into N equal parts
				 using MFA Potts */

  free(nod);
  free(err);

  return;

}    /* end of main() */


void  mfa_potts_worker(double tmp, double alpha)
{

/*** allocate dynamic space for vectors v[] and s[] and matrix t[][] ***/

     t = calloc(ne*ne,sizeof(int));
     v = calloc(ne,sizeof(double));


/*** set up the weight matrix of connecting interfaces between elements ***/
/* the matrix consists of 0's for no connection and interface lengths */

weight_matrix();


/*** Receive initialized v[] vector from Root Processor ***/


rec_v();


/****** Perform Annealing of the Mean Field Theory  ******/

anneal_worker(tmp, alpha);
	
/*****   End of Annealing    *****/


} /* end of mfa */




void   weight_matrix()
{
int     a, b, i, j, k, m, n, counter, x[3],xx[3], nd[2];

      if(ne > 0)
    {
      for(i = 0; i < ne; i++)
      {
	for(j=0; j<3; j++)
	  x[j] = nod[4*i+j];
	for(k=0; k<ne; k++)
	{
	  counter=0;
	  if (k == i) t[k+i*ne] = 0;
	  else
	  {
	    for(m=0; m<3; m++)
	      xx[m] = nod[4*k+m];
	    for(a=0; a<3; a++)
	    {
	      for(b=0; b<3; b++)
	      {
		if(x[a] == xx[b])
		{
		  nd[counter] = x[a];
		  counter++;
		}
		if(counter==2) break;
	      }
	      if(counter == 2)break;
	    }
	  }
	  if(counter == 2)
	   t[k+i*ne] = 1;

	  else t[k+i*ne] = 0;
	}
      }
    }

}  /*  end of weight_matrix()  */


void   rec_v()
{
int    i,j,msl_vinit;
double  vinit;

msl_vinit = sizeof(double);
	
   for(i=0; i<ne; i++)
   {
	if(proc_id == num_proc)
	{
	      chan_in_message(msl_vinit, &vinit, in_p[1]);
	      v[i] = vinit;
	}
	else
	{
	   for(j=0; j<num_proc-proc_id; j++)
	   {
		  chan_in_message(msl_vinit, &vinit, in_p[1]);
		  v[i] = vinit;
		  chan_out_message(msl_vinit, &vinit, out_p[0]);
	   }
	      chan_in_message(msl_vinit, &vinit, in_p[1]);    
	      v[i] = vinit;
	}
   }
}



void  anneal_worker(double tmp, double alpha)
{
double  sum, *h, *vnew, test;
int     i, j, msl_vnew, msl_h;

    msl_h = num_proc*sizeof(double);
    msl_vnew = num_proc*sizeof(double);

    h = calloc(num_proc,sizeof(double));
    vnew = calloc(num_proc,sizeof(double));

while (1)
{
   for(i=0; i<ne; i++)
   {
	sum = 0.0;
	for (j=0; j<ne; j++)
	{
	    if(j != i)
	    {  
		test = v[j];
		if(choice) sum += ((double)t[j+i*ne] - (alpha * err[j]))
				   * v[j];
		 else sum += ((double)t[j+i*ne] - alpha) * v[j];
	    }      
	}   /* end of j for loop */

  
      
	/* send the sum to the root */

	if(proc_id == num_proc)
	{
	     h[num_proc-proc_id] = sum;    
	     chan_out_message(msl_h, h, out_p[1]);
	}
	else
	{
	     chan_in_message(msl_h, h, in_p[0]);
	     h[num_proc-proc_id] = sum;    
	     chan_out_message(msl_h, h, out_p[1]);
	}
	 
	 /* receive the new value for v */

	if(proc_id == num_proc)
	    chan_in_message(msl_vnew, vnew, in_p[1]);
	else
	{
		   chan_in_message(msl_vnew, vnew, in_p[1]);
		   chan_out_message(msl_vnew, vnew, out_p[0]);
	}
		     v[i] = vnew[num_proc-proc_id];  
   }   /* end of i for loop */


} /* end of while loop */

}  /*  end of anneal()  */


